with open('ordinalEncoder.pkl', 'rb') as f:
    ordinal_encoder = pickle.load(f)

CATEGORICAL_COLS = {{ columns }}
{{ test_dataset }}[CATEGORICAL_COLS] = ordinal_encoder.transform({{ test_dataset }}[CATEGORICAL_COLS])